use std::collections::HashMap;
use std::path::PathBuf;
use chrono::NaiveDate;
use crate::models::StrMessage;

fn header(title: &str) -> String{
    format!(r#"<!DOCTYPE html>
        <html>
            <head>
                <title>{title}</title>
                <meta http-equiv='Permissions-Policy' content='interest-cohort=()'/>
                <meta name='viewport' content='width=device-width, initial-scale=1.0, maximum-scale=1.0,user-scalable=0' />
                <link rel='icon' href='data:image/svg+xml,<svg xmlns=%22http://www.w3.org/2000/svg%22 viewBox=%220 0 100 100%22><text y=%22.9em%22 font-size=%2290%22>📧</text></svg>'>
                <meta name="description" content="{title}"/>
            </head>
        <body>
            <div>
                <b>sort by</b> [<a href='authors.html'>author</a>] [<a href='dates.html'>dates</a>] [<a href='subjects.html'>subjects</a>]
            </div>
            <ul>
        "#)
}

fn footer() -> String{
    r#"
        </ul>
        </body>
        </html>
    "#.to_string()
}

pub struct AuthorIndex{
    author_indexes: HashMap<String, Vec<StrMessage>>
}

pub struct DateIndex{
    date_indexes: HashMap<String, Vec<StrMessage>>
}

pub struct SubjectIndex{
    subject_indexes: HashMap<String, Vec<StrMessage>>
}

impl AuthorIndex{
    // default constructor
    pub fn new() -> Self{
        AuthorIndex {
            author_indexes: HashMap::new(),
        }
    }

    // adds an email hash to the hashmap with a author as a key
    pub fn add_indexed_email(&mut self, message: StrMessage){
        self.author_indexes.entry(message.from.address.clone())
            .or_insert_with(Vec::new)
            .push(message.clone());
    }

    // gets a list of emails by author
    pub fn get_by_author(&mut self, author: String) -> Option<&Vec<StrMessage>> {
        self.author_indexes.get(&author)
    }

    // gets a list of all authors
    pub fn get_authors(&mut self) -> Vec<String>{
        self.author_indexes.keys().cloned().collect()
    }

    pub fn to_authors_html(&mut self, base_path: PathBuf){
        let path = base_path.join("authors.html");
        let mut res = String::new();
        res.push_str(&*header("Authors"));

        // change key from email to name
        for author in self.get_authors(){
            let name = self.get_by_author(author.clone()).unwrap()[0].clone().from.name;
            if name != None {
                if let Some(emails) = self.author_indexes.remove(&author) {
                    self.author_indexes.insert(name.unwrap().clone(), emails);
                }
            }
        }

        let mut keys = self.get_authors();
        keys.sort();
        for author in keys {
            res.push_str(&format!("<li><b>{}</b><ul>", author));
            for message in self.get_by_author(author.clone()).unwrap() {
                res.push_str(&format!("<li><a href={:?}>{:?}</a></li>",
                                      message.path.clone(),
                                      message.subject.clone()
                ));
            }
            res.push_str("</ul></li>");
        }

        res.push_str(&*footer());

        std::fs::write(path.clone(), res).unwrap();
    }
}

impl DateIndex{
    // default constructor
    pub fn new() -> Self{
        DateIndex {
            date_indexes: HashMap::new(),
        }
    }

    // adds an email hash to the hashmap with a author as a key
    pub fn add_indexed_email(&mut self, message: StrMessage){
        let date = message.clone().date;
        self.date_indexes.entry(date.year.to_string() + "-" + &*date.month.to_string() + "-" + &*date.day.to_string())
            .or_insert_with(Vec::new)
            .push(message.clone());
    }

    // gets a list of emails by date
    pub fn get_by_date(&mut self, date: String) -> Option<&Vec<StrMessage>> {
        self.date_indexes.get(&date)
    }

    // gets a list of all dates
    pub fn get_dates(&mut self) -> Vec<String>{
        self.date_indexes.keys().cloned().collect()
    }

    pub fn to_dates_html(&mut self, base_path: PathBuf){
        let path = base_path.join("dates.html");
        let mut res = String::new();
        res.push_str(&*header("Dates"));

        let mut keys = self.get_dates();
        let keys_converted: Result<Vec<NaiveDate>, chrono::ParseError> = keys.clone().into_iter()
            .map(|date_str| NaiveDate::parse_from_str(&date_str, "%Y-%-m-%-d"))
            .collect();

        keys = match keys_converted {
            Ok(mut dates) => {
                dates.sort();
                dates.reverse();
                dates.iter().map(|date| date.format("%Y-%-m-%-d").to_string()).collect()
            },
            Err(_e) => {
                keys.sort();
                keys.reverse();
                keys
            }
        };
        
        for _date in keys {
            res.push_str(&format!("<li><b>{}</b><ul>", _date));
            for message in self.get_by_date(_date.clone()).unwrap() {
                res.push_str(&format!("<li><a href={:?}>{:?}</a></li>",
                                      message.path.clone(),
                                      message.subject.clone()
                ));
            }
            res.push_str("</ul></li>");
        }

        res.push_str(&*footer());

        std::fs::write(path.clone(), res).unwrap();
    }
}

impl SubjectIndex{
    // default constructor
    pub fn new() -> Self{
        SubjectIndex {
            subject_indexes: HashMap::new(),
        }
    }

    // adds an email hash to the hashmap with a author as a key
    pub fn add_indexed_email(&mut self, message: StrMessage){
        self.subject_indexes.entry(message.clone().subject)
            .or_insert_with(Vec::new)
            .push(message.clone());
    }

    // gets a list of emails by subject
    pub fn get_by_subject(&mut self, subject: String) -> Option<&Vec<StrMessage>> {
        self.subject_indexes.get(&subject)
    }

    // gets a list of all subjects
    pub fn get_subjects(&mut self) -> Vec<String>{
        self.subject_indexes.keys().cloned().collect()
    }

    pub fn to_subjects_html(&mut self, base_path: PathBuf){
        let path = base_path.join("subjects.html");
        let mut res = String::new();
        res.push_str(&*header("Subjects"));

        let mut keys = self.get_subjects();
        keys.sort();
        for subject in keys {
            res.push_str(&format!("<li><b>{}</b><ul>", subject));
            for message in self.get_by_subject(subject.clone()).unwrap() {
                res.push_str(&format!("<li><a href={:?}>{:?}</a></li>",
                                      message.path.clone(),
                                      message.subject.clone()
                ));
            }
            res.push_str("</ul></li>");
        }

        res.push_str(&*footer());

        std::fs::write(path.clone(), res).unwrap();
    }
}