#![forbid(unsafe_code)]
use anyhow::Result;
use mail_parser::Message;
use std::collections::{HashSet};
use std::fs;
use std::path::{Path, PathBuf};

use models::*;

use config::{Config, INSTANCE};
mod arg;
mod config;
mod maildir;
mod models;
mod templates;
mod threading;
mod util;
mod indexes;
mod imap;

use std::ffi::{OsStr, OsString};
use crate::indexes::{AuthorIndex, DateIndex, SubjectIndex};

const ATOM_ENTRY_LIMIT: usize = 100;

// stole it from the internet
pub fn append_ext(ext: impl AsRef<OsStr>, path: &PathBuf) -> PathBuf {
    let mut os_string: OsString = path.into();
    os_string.push(".");
    os_string.push(ext.as_ref());
    os_string.into()
}

impl Lists {
    fn write_lists(&mut self, author_indexes: &mut AuthorIndex, date_indexes: &mut DateIndex, subject_indexes: &mut SubjectIndex) {
        std::fs::create_dir_all(&self.out_dir).ok();
        write_if_changed(&self.out_dir.join("style.css"), include_bytes!("style.css"));
        write_if_changed(&self.out_dir.join("floatingmenu.js"), include_bytes!("floatingmenu.js"));
        let base_path = self.out_dir.join("index");
        if Config::global().include_html {
            write_if_changed(&base_path.with_extension("html"), self.to_html());
        }
        if Config::global().include_gemini {
            write_if_changed(&base_path.with_extension("gmi"), self.to_gmi());
        }
        for list in &mut self.lists {
            list.persist(author_indexes, date_indexes, subject_indexes);
        }
    }
}

/// Writes the given data to the given path iff the data is not
/// identical to the file contents.
///
/// Returns true if it wrote to the file.
fn write_if_changed<T: AsRef<[u8]>>(path: &PathBuf, data: T) -> bool {
    // TODO: use checksum / cache
    if let Ok(d) = std::fs::read(path) {
        if d == data.as_ref() {
            return false;
        }
    }

    std::fs::write(path, data).unwrap();
    true
}

impl List {
    fn persist(&mut self, author_indexes: &mut AuthorIndex, date_indexes: &mut DateIndex, subject_index: &mut SubjectIndex) {
        self.write_threads(author_indexes, date_indexes, subject_index);
    }
    fn write_index(&self) {
        fn write(elems: Vec<String>, out_dir: &Path, ext: &str) {
            for (n, text) in elems.iter().enumerate() {
                let index = if n == 0 {
                    out_dir.join("index")
                } else {
                    out_dir.join(format!("{}-{}", "index", n + 1))
                };
                write_if_changed(&index.with_extension(ext), text);
            }
        }

        if Config::global().include_gemini {
            write(self.to_gmi(), &self.out_dir, "gmi");
        }
        if Config::global().include_html {
            write(self.to_html(), &self.out_dir, "html");
        }
    }

    // Used with atom
    fn get_recent_messages(&self) -> Vec<StrMessage> {
        let mut out = Vec::new();
        let mut msgs: Vec<&threading::Msg> = self.thread_idx.threads.iter().flatten().collect();
        msgs.sort_by(|a, b| a.time.cmp(&b.time));
        msgs.reverse();
        for m in msgs.iter().take(ATOM_ENTRY_LIMIT) {
            let data = std::fs::read(&m.path).unwrap();
            let msg = StrMessage::new(&Message::parse(&data).unwrap());
            out.push(msg);
        }
        out
    }

    fn write_threads(&mut self, author_indexes: &mut AuthorIndex, date_indexes: &mut DateIndex, subject_indexes: &mut SubjectIndex) {
        let mut files_written: HashSet<PathBuf> = HashSet::new();
        let out_dir = self.out_dir.clone();
        // creating directories for threads and messages
        let thread_dir = out_dir.join("threads");
        let message_dir = out_dir.join("messages");
        std::fs::create_dir_all(&thread_dir).ok();
        std::fs::create_dir_all(&message_dir).ok();
        
        // iterating through each thread
        for thread_ids in &self.thread_idx.threads {
            let mut thread = Thread::new(thread_ids, &self.config.name, &self.config.email);
            
            let base_path = thread_dir.join(thread.messages[0].hash.clone()); // path without extension

            let summary = ThreadSummary {
                message: thread.messages[0].clone(),
                reply_count: (thread.messages.len() - 1) as u64,
                last_reply: thread_ids[thread_ids.len() - 1].time.clone(),
            };
            
            let mut hash_path = PathBuf::from(out_dir.clone().file_name().unwrap().to_str().unwrap().to_string());
            hash_path = hash_path.clone().join("threads").join(thread.messages[0].hash.clone());
            hash_path = append_ext("html", &hash_path.clone());
            for msg in &mut thread.messages {                
                msg.set_url(self, msg.hash.clone());
                msg.set_path(hash_path.clone().to_str().unwrap().to_string() + "#" + &*msg.hash.clone());
                
                // indexing
                author_indexes.add_indexed_email(msg.clone());
                date_indexes.add_indexed_email(msg.clone());
                subject_indexes.add_indexed_email(msg.clone());
            }
            self.thread_topics.push(summary);
            if Config::global().include_html {
                let html = append_ext("html", &base_path);
                write_if_changed(&html, thread.to_html(out_dir.clone()));
                files_written.insert(html);
            }
            let xml = append_ext("xml", &base_path);
            write_if_changed(&xml, thread.to_xml());
            files_written.insert(xml);
            if Config::global().include_gemini {
                let gmi = append_ext("gmi", &base_path);
                write_if_changed(&gmi, thread.to_gmi());
                files_written.insert(gmi);
            }

            for msg in thread.messages {                
                let eml = append_ext("mbox", &message_dir.join(msg.hash.clone()));
                write_if_changed(&eml, &msg.export_mbox());
                files_written.insert(eml);
            }
        }
        self.thread_topics
            .sort_by(|a, b| a.last_reply.cmp(&b.last_reply));
        self.thread_topics.reverse();
        self.recent_messages = self.get_recent_messages();

        // Remove deleted stuff
        for dir in vec![message_dir, thread_dir] {
            for entry in fs::read_dir(&dir).unwrap() {
                match entry {
                    Ok(e) => {
                        if !files_written.contains(&e.path()) {
                            fs::remove_file(&e.path()).ok();
                        }
                    }
                    Err(_) => continue,
                }
            }
        }
        //
        self.write_index();
    }
}

#[tokio::main]
async fn main() {
    // creating empty hashMaps for indexing
    let mut author_indexes = AuthorIndex::new();
    let mut date_indexes = DateIndex::new();
    let mut subject_indexes = SubjectIndex::new();
    
    // reading args
    let args = arg::Args::from_env();
    let maildir = &args.positional[0];
    let mut config = Config::from_file(&args.config).expect("could not read config");
    // Default to both true if both absent
    if !args.include_gemini && !args.include_html {
        config.include_gemini = true;
        config.include_html = true;
    } else {
        config.include_gemini = args.include_gemini;
        config.include_html = args.include_html;
    }
    config.out_dir = args.out_dir.clone();

    // downloading new emails
    imap::download_email_from_imap(
        PathBuf::from(maildir.clone()),
        config.imap_domain.clone(),
        config.username.clone(),
        config.password.clone()
    ).await.expect("Cannot parse config for imap_domain or username or password");

    INSTANCE.set(config).unwrap();
    
    // creating a vector of lists
    let mut lists = Lists {
        lists: vec![],
        out_dir: Config::global().out_dir.clone(),
    };
    // reading the folder with lists
    for maildir in std::fs::read_dir(maildir)
        .expect("could not read maildir")
        .filter_map(|m| m.ok())
    {
        let dir_name = maildir.file_name().into_string().unwrap();
        // if this is not a list - continue
        if dir_name.starts_with('.') || ["cur", "new", "tmp"].contains(&dir_name.as_str()) {
            continue;
        }

        let mut list = threading::ThreadIdx::new();
        // get all mails from maildir
        let mails = maildir::list_all(maildir.path())
            .expect("maildir configured incorrectly")
            .filter_map(Result::ok);

        for mail in mails {
            // getting the content of the mail
            let path = mail.path().to_path_buf();
            let data = std::fs::read(&path).expect("could not read mail");

            if let Some(mail) = mail_parser::Message::parse(&data) {
                list.add_email(&mail, path);
            } else {
                println!("Could not parse message {:?}", path);
            }
        }

        list.finalize();
        lists.add(list, &dir_name);
    }
    lists.write_lists(&mut author_indexes, &mut date_indexes, &mut subject_indexes);
    
    // write indexed emails to htmls
    let p = args.out_dir.to_path_buf().clone();
    author_indexes.to_authors_html(p.clone());
    date_indexes.to_dates_html(p.clone());
    subject_indexes.to_subjects_html(p.clone());
}
